# Package: CPP
# Version: 1.0
# Filename: _CPP_annotation.praat
#
# Author: Santitham Prom-on, Yi Xu
# Last updated: 16 July 2016
#
# Objective: annotate input speech files iteratively. This process includes 
# pulse marking, pitch extraction, and boundary identification

## Macro ##
taskEdit = 1
taskResume = 2
taskCreate = 3
#

##### Input Module - Begin #####

#Select the working directory
directory$ = chooseDirectory$ ("Choose a directory to annotate")
if directory$ = ""
	#[EXIT_00] Choose to not Continue
	exit
endif

#Check whether the selected directory contain .wav or .WAV files
Create Strings as file list... FileList 'directory$'/*.wav
numberOfFiles = Get number of strings
if !numberOfFiles
	select Strings FileList
	Remove
	Create Strings as file list... FileList 'directory$'/*.WAV
	numberOfFiles = Get number of strings
endif

#[Error checking] No files in the directory	
if numberOfFiles = 0
	#[EXIT_A01] Sound files not found
	exit Error Code A01: No sound files in the directory
else
	Write to raw text file... 'directory$'/FileList.txt
endif

Create Table with column names... f0check 'numberOfFiles' Name Flag

#Read annotation files
Create Strings as file list... annot_list 'directory$'/*.intervals
n_annot = Get number of strings
Remove

# Main option window
beginPause("Annotation - Options")
	comment("Select the annotation task")
	optionMenu("Task:", 1)
		option("Edit annotation")
		option("Process all files without pause")
		integer("input_File_No", 1)
		word("TextGrid_extension", ".TextGrid")
		word("Sound_file_extension", ".wav")
		boolean("Edit F0 analysis options",0)
	boolean("Inspect pulse",0)
clicked = endPause("Cancel","Start",2,1)

#[EXIT_00] Choose to not Continue
if clicked = 1
	exit
endif

# If users want to edit the F0 analysis option
	
if edit_F0_analysis_options
	#Prompt the window to display analysis options
	beginPause ("Edit F0 Analysis Options")
		integer("left F0 range (Hz)",75)
		integer("right F0 range (Hz)", 500)
		integer("Number of point per interval",10)
		integer("F0 sample rate (Hz)",100)
		boolean("Set initial time for normf0 to 0",1)
	clicked = endPause("Cancel","Continue",2,1)
	
	#[EXIT_00] Choose to not continue
	if clicked = 1
		exit
	endif
	
	npoints = number_of_point_per_interval
	minf0 = left_F0_range
	maxf0 = right_F0_range
else
	#Edit F0 analysis option is not selected. Set all options to default values
	npoints = 10
	minf0 = 75
	maxf0 = 500
	f0_sample_rate = 100
	set_initial_time_for_normf0_to_0 = 1
endif

# for each file (from input_File_No to numberOfFiles), call Labeling
current_file = input_File_No
while current_file <= numberOfFiles
	select Strings FileList
	fileName$ = Get string... current_file
	name$ = fileName$ - ".wav" - ".WAV"
	call Labeling 'fileName$'
	if task = 2
		current_file += 1
	endif
endwhile

call ReportError

if cflag = 0
	printline === Error has occurred: Annotation incomplete ===
else
	printline === Annotation completed ===
endif

select Strings FileList
plus Table f0check
Remove
#===End of Main===
#####

##########Other Procedures##########

procedure Labeling file_name$
	#Prepare the files necessary to annotation
	Read from file... 'directory$'/'file_name$'
	Read from file... 'directory$'/'file_name$'
	name$ = file_name$ - ".wav" - ".WAV"
	labelfile$ = name$+"'textGrid_extension$'"
	textgridfile$ = name$+".TextGrid"
	if fileReadable ("'directory$'/'labelfile$'")
		Read from file... 'directory$'/'labelfile$'
	elsif fileReadable (textgridfile$)
		Read from file... 'directory$'/'name$'.TextGrid
	else
		To TextGrid... "interval point" point
	endif
	plus Sound 'name$'
	Edit
	
	pulsefile$ = name$+".pulse"
	if fileReadable ("'directory$'/'pulsefile$'")
		Read from file... 'directory$'/'pulsefile$'
	else
		select Sound 'name$'
		To PointProcess (periodic, cc)... minf0 maxf0
		Rename... 'name$'
	endif

	if inspect_pulse
		plus Sound 'name$'
		Edit
	endif

	if task = 1
		beginPause ("Press Done to exit")
			jump_to = current_file
			integer ("Jump to", jump_to)
		clicked2 = endPause ("Stop", "Back", "Next", "Jump", "Done", 3, 1)
		if clicked2 = 1
			nocheck plus TextGrid 'name$'
			nocheck plus PointProcess 'name$'
			Remove
			exit
		elsif clicked2 = 2
			current_file = if current_file > 1 then current_file-1 else 1 fi
		elsif clicked2 = 3
			if current_file < numberOfFiles
				current_file += 1
			else
				call Save 'directory$' 'name$'
				echo Current file number is 'current_file'.
				select Sound 'name$'
				Remove
				exit
			endif
		elsif clicked2 = 4
			current_file = if jump_to < numberOfFiles then jump_to else numberOfFiles fi
		elsif clicked2 = 5
			call Save 'directory$' 'name$'
			echo Current file number is 'current_file'.
			select Sound 'name$'
			Remove
			exit
		endif			
	endif
		

	call Save 'directory$' 'name$'
	
	select Sound 'name$'
	Remove
endproc


###Clean all files generated for the target file
procedure deleteAuxFiles name$
	if fileReadable("'directory$'/'name$'.annotation")
		deleteFile("'directory$'/'name$'.annotation")
	endif
	if fileReadable("'directory$'/'name$'.annotation_short")
		deleteFile("'directory$'/'name$'.annotation_short")
	endif
	if fileReadable("'directory$'/'name$'.pulse")
		deleteFile("'directory$'/'name$'.pulse")
	endif
endproc

###Save the auxiliary files related to name$###
procedure Save directory$ name$
	select PointProcess 'name$'
	npulses = Get number of points
	for n from 2 to npulses
		time1 = Get time from index... n-1
		time2 = Get time from index... n
		if time2 = time1
			Remove point... n
		endif
	endfor
	Write to short text file... 'directory$'/'name$'.pulse
	maxperiod = 1/minf0
	To PitchTier... maxperiod
	call Trimf0

	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	sampleStart = Get starting point... 1 1
	sampleEnd = Get end point... 1 nintervals
	call Smooth 'name$' "smoothf0" npoints
	
	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	interval = 0
	found_interval = 0
	nrows = 0
	for m from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 m
		if not label$ = ""
			found_interval = 1
		endif
	endfor
	if found_interval
		call Normalization smoothf0
		
		call CheckF0
		
		select TableOfReal normactuf0
		Write to headerless spreadsheet file... 'directory$'/'name$'.actutimenormf0
		plus TableOfReal normf0
		plus TableOfReal normactuf0
		plus PitchTier smoothf0
		Remove
		
		select PitchTier 'name$'
		Remove
	endif
	
	select TextGrid 'name$'
	Write to short text file... 'directory$'/'name$'.TextGrid

	Down to Table... no 4 yes no
	Save as tab-separated file... 'directory$'/'name$'.intervals
	Remove
	
	select TextGrid 'name$'
	plus Sound 'name$'
	plus PointProcess 'name$'
	Remove
endproc


procedure Trimf0
	maxbump = 0.01
	maxedge = 0.0
	maxgap = 0.033
	n = Get number of points
	
	first = Get value at index... 1
	second = Get value at index... 2
	penult = Get value at index... n-1
	last = Get value at index... n
	tfirst = Get time from index... 1
	tlast = Get time from index... n
	for k from 1 to 3
		call Trim
	endfor
	Remove point... 1
	Add point... tfirst second + (first-second) / 1000
	Remove point... n
	Add point... tlast penult + (last-penult) / 1000
endproc

procedure Trim
	for i from 2 to n-1
		tleft = Get time from index... i-1
		tmid = Get time from index... i
		tright = Get time from index... i+1
		gap1 = tmid - tleft
		gap2 = tright - tmid
		left = Get value at index... i-1
		mid = Get value at index... i
		right = Get value at index... i+1
		diff1 = mid - left
		diff2 = mid - right
		if diff1 > maxbump and diff2 > maxedge and gap1 < maxgap and gap2 < maxgap
		... or diff2 > maxbump and diff1 > maxedge and gap1 < maxgap and gap2 < maxgap
			Remove point... i
			Add point... tmid left+(tmid-tleft)/(tright-tleft)*(right-left)
		endif
		if diff1 > maxbump and gap2 >= maxgap
			Remove point... i
			Add point... tmid left + maxbump
		endif
		if diff2 > maxbump and gap1 >= maxgap
			Remove point... i
			Add point... tmid right + maxbump
		endif

		diff1 = left - mid
		diff2 = right - mid
		if diff1 > maxbump and diff2 > maxedge and gap1 < maxgap and gap2 < maxgap
		... or diff2 > maxbump and diff1 > maxedge and gap1 < maxgap and gap2 < maxgap
			Remove point... i
			Add point... tmid left+(tmid-tleft)/(tright-tleft)*(right-left)
		endif
		if diff1 > maxbump and gap2 >= maxgap
			Remove point... i
			Add point... tmid left - maxbump
		endif
		if diff2 > maxbump and gap1 >= maxgap
			Remove point... i
			Add point... tmid right - maxbump
		endif
	endfor
endproc

procedure Normalization PitchTier$
	select PitchTier 'PitchTier$'
	lasttime = Get finishing time
	
	# create empty PitchTiers named "normactutimef0" and "normactuf0"
	Create PitchTier... normactutimef0 0 lasttime
	Create TableOfReal... normf0 1 2
	Set column label (index)... 1 NormalizedTime
	Set column label (index)... 2 F0
	Create TableOfReal... normactuf0 1 2
	Set column label (index)... 1 ActualTime
	Set column label (index)... 2 F0
	
	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	interval = 0
	found_interval = 0
	nrows = 0
	# for each interval
	for m from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 m
		if not label$ = ""
			start = Get starting point... 1 m
			end = Get end point... 1 m
			select PitchTier 'PitchTier$'
			index1 = Get high index from time... start
			index2 = Get low index from time... end
			if found_interval = 0
				found_interval = 1
				select PitchTier 'PitchTier$'
				firstime = start
			endif
			call Normalize
			interval = interval + 1
		endif
	endfor
	select TableOfReal normf0
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
	select TableOfReal normactuf0
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
	select PitchTier normactutimef0
	Remove
endproc

procedure Normalize
	duration = end - start

	for i from index1 to index2
		select PitchTier 'PitchTier$'
		time = Get time from index... i
		f0 = Get value at index... i
		select PitchTier normactutimef0
		Add point... (time-start)/duration+interval f0
	endfor
	
	for x from 1 to npoints
		normtime = x / npoints
		select PitchTier normactutimef0
		f0 = Get value at time... normtime+interval
		select TableOfReal normf0
		nrows = Get number of rows
		Set value... nrows 1 x+interval*npoints
		Set value... nrows 2 f0
		Set row label (index)... nrows 'label$'
		Insert row (index)... nrows + 1
		select TableOfReal normactuf0
		nrows = Get number of rows
		Set value... nrows 1 normtime*duration+start
		if set_initial_time_for_normf0_to_0
			Set value... nrows 1 normtime*duration+start-firstime
		endif
		Set value... nrows 2 f0
		Set row label (index)... nrows 'label$'
		Insert row (index)... nrows + 1
	endfor
endproc

procedure Smooth curve_in$ curve_out$ width
	Create PitchTier... 'curve_out$' sampleStart sampleEnd
	
	for j from 1 to width							; make a triangular window of size = width 
		if j < width / 2 + 0.5
			weight'j' = j
		else 
			weight'j' = width - j + 1
		endif
	endfor

	select PitchTier 'curve_in$'
	smooth_end = Get number of points

	for i from 1 to width / 2						; smooth initial points: 0 to width/2 - 1 
	    n = 0.0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... i
	    for j from 1 to width/2 + i					; window size = width/2 to width - 1
	    	j = 'j:0'
	    	rawsample = Get value at index... j
	    	index = width / 2 + j - i
	    	index = 'index:0'
			smoothsample += weight'index' * rawsample
			n += weight'index'
	    endfor
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
	endfor
	
	for i from width/2 to smooth_end - width/2				; smooth from width/2 to end-width/2
		n = 0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... i
		for j from 1 to width
	    	rawsample = Get value at index... i-width/2+j
			smoothsample += weight'j' * rawsample
			n += weight'j'
		endfor
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
	endfor
	
	i = width/2
	while i > 0										; smooth final points: end - width/2 to end
		n = 0.0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... smooth_end-i
		j = width/2 + i
		j = 'j:0'
		while j > 1									; window size = width - 1 to width/2 
	    	rawsample = Get value at index... smooth_end-j
	    	index = width/2+i-j + 1
	    	index = 'index:0'
			smoothsample += weight'index' * rawsample
			n += weight'index'
			j -= 1
		endwhile
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
		i -= 1
	endwhile
endproc

procedure CheckF0
	flag = 1
	select TableOfReal normf0
	nrows = Get number of rows
	for i from 1 to nrows
		fx = Get value... i 2
		if fx = undefined
			flag = 0
		endif
	endfor
	
	select TableOfReal normactuf0
	nrows = Get number of rows
	for i from 1 to nrows
		fx = Get value... i 2
		if fx = undefined
			flag = 0
		endif
	endfor
		
	select Table f0check
	Set string value... current_file Name 'name$'
	Set numeric value... current_file Flag flag
;	Remove
endproc

procedure ReportError
	cflag = 1
	select Table f0check
	nrows = Get number of rows
	for i from 1 to nrows
		select Table f0check
		flag = Get value... i Flag
		name$ = Get value... i Name
		if flag = 0
			printline Error: undefined value at file: 'name$'
			cflag = 0
		endif
	endfor
endproc
